;; Generated package description from diff-hl.el
(define-package "diff-hl" "1.8.5" "Highlight uncommitted changes using VC" '((cl-lib "0.2") (emacs "24.3")) :url "https://github.com/dgutov/diff-hl" :keywords '("vc" "diff"))
